/*
 * DetectAgent.java
 *
 * Created on 1 dicembre 2004, 19.41
 */

package sid.agent;


import alice.tucson.api.*;
import alice.logictuple.*;

import alice.tuprolog.Prolog;
import alice.tuprolog.Theory;
import alice.tuprolog.Term;
import alice.tuprolog.Number;
import alice.tuprolog.Int;
import alice.tuprolog.Struct;
import alice.tuprolog.SolveInfo;


import java.io.*;

/**
 *agente rilevatore di mine
 * @author  matteo
 */
public class DetectAgent extends sid.agent.RobotAgent {
    
    
    /**
     *percorso per le teorie per il motore prolog dell'agente
     */
     private final String thPathCommon=sid.gui.SimulationGUI.userPath+sid.gui.SimulationGUI.fileSeparator+"theories"+sid.gui.SimulationGUI.fileSeparator+"commonTheory.pl";
     
     private final String thPathAgent=sid.gui.SimulationGUI.userPath+sid.gui.SimulationGUI.fileSeparator+"theories"+sid.gui.SimulationGUI.fileSeparator+"agentDetectTheory.pl";
       
     
     //direzione preferita di movimento casuale: 1 se preferiscono muoversi verso "ovest", 2 se preferiscono muoversi verso "est"
     private int movDir;
    /**
     *riferimento alla matrice usata per aggiornare l'interfaccia grafica del campo minato
     */
    private int[][] matrix;
    
    /** 
     *costruttore
     *@param x la coordinata x che relativa alla posizione iniziale dell' Agent
     *@param y la coordinata y relativa alla posizione iniziale dell' Agent
     *@param matrix riferimento alla matrice usata per aggiornare l'interfaccia grafica del campo minato
     *@param name il nome del robot agent, ovvero il suo identificativo
     *
     */
    public DetectAgent(int x, int y,int[][] matrix, AgentId name) throws Exception {
        
        super(x,y,name);
        
        this.matrix=matrix;
        
        //scelta casuale secondo distribuzione uniforme della direzione preferita casuale di movimento a parit di celle adiacenti libere
        this.movDir=( Math.random()<0.5 ? 1 : 2);
        
        //System.out.println("mov random type: "+this.movDir);
      
    }//costructor
    
    /**
     *definizione del body dell'agente rilevatore di mine
     *
     */
    protected void body(){
        
        try{
            
            
            
            
            
            //creo gli id dei due centri di tuple principali che utilizzer

            //creo il tuplecenter principale che rappresenta il campo minato
            TupleCentreId tcMineField=new TupleCentreId("mineField");


            //creo il tuplecenter principale che rappresenta il campo minato
            TupleCentreId tcRadioChannel=new TupleCentreId("radioChannel");
        
            
            
            System.out.println("Agente detect "+this.getId()+" in esecuzione");
                
            
            //creo un nuovo motore prolog per l'attivita di pianificazione e deliberazione dell'agente
            Prolog engine=new Prolog();
            
            
            //file con teoria prolog di utilit generale
            //FileInputStream thFileCommon=new FileInputStream(thPathCommon);
            
            //teoria creata a partire dal file precedente
            //Theory commonTheory=new Theory(thFileCommon);
            
            //file con teoria prolog specifica dell'agente
            FileInputStream thFileAgent=new FileInputStream(thPathAgent);
            
            //teoria creata a partire dal file precedente
            Theory agentTheory=new Theory(thFileAgent);
            
            //aggiunta teoria all'engine
            
            //engine.addTheory(commonTheory);
            engine.addTheory(agentTheory);
                     
                           
            
            
             //viene impostata la posizione dell'agente anche nel centro di tuple che rappresenta l'ambiente ovvero il campo minato
             //cn.out(tcMineField,new LogicTuple("cell",new Value(this.getX()),new Value(this.getY()),new Value("agent",new Value("detect"),new Value(this.getId().toString()))));
             
            
            
            //appena creato l'agente inserisce la sua posizione nel centro di tuple che rappresenta il canale radio: comunica agli altri la sua posizione
            out(tcRadioChannel,new LogicTuple("agent",new Value(this.getX()),new Value(this.getY()),new Value("detect"),new Value(this.getId().toString())));
            
            //aggiungo la posizione iniziale dell'agente alla teoria dell'engine e il numero del turno
            engine.addTheory(new Theory("myPos("+this.getX()+","+this.getY()+"). \n turnNum(0). \n visited("+this.getX()+","+this.getY()+",none,[])."));
            
            //l'agente recupera le dimensioni massime del campo minato
            Var m=new Var("MaxDim");
            
            rdp(tcMineField,new LogicTuple("maxDim",new Value("x"),m));
            
            int xDim=m.getVarValue("MaxDim").intValue();

            rdp(tcMineField,new LogicTuple("maxDim",new Value("y"),m));
            
            int yDim=m.getVarValue("MaxDim").intValue();
            
            //numero massimo di turni che l'agente attende senza muoversi
            int waitTurn=5;
            
             
            
            
            
            //ciclo dell'agente
            while(true){
               
                Var agentNo=new Var("AgNum");
                Var turnNum=new Var("TurnNum");
                
                //lista (inizialmente vuota) contentente tutte le tuple che si trovano attorno alla posizione attuale dell'agente
                //Struct list=new Struct();
                
                //System.out.println(list.toString());
                
                //l'agente aspetta che sia il proprio turno: attesa per il token turn(TurnNum,AgentNum)
                in(tcMineField,new LogicTuple("turn",turnNum,agentNo));
                
                System.out.println("************turno numero "+turnNum.getVarValue("TurnNum").intValue()+" di DetectAgent "+this.getId()+"; agente no: "+agentNo.getVarValue("AgNum"));
                
                
                //l'agente controlla che non sia arrivata la fine della simulazione
                LogicTuple endSim=rdp(tcMineField,new LogicTuple("endSim"));
                
                //se  arrivata la fine della simulazione allora si esce da ciclo e l'agente termina
                if(endSim!=null) break;
                
                /*
                //solo per testing: celle visitate dall'agente
                SolveInfo test=engine.solve("visited(X,Y,Prev,Next).");
                
                System.out.println(""+test);
                
                while(engine.hasOpenAlternatives())
                    System.out.println(""+engine.solveNext());
                
                test=engine.solve("myPos(X,Y).");
                
                System.out.println(""+test);
                
                while(engine.hasOpenAlternatives())
                    System.out.println(""+engine.solveNext());
                
                */
                   
                
                //*****************************************************************************************************************************************************************
                //l'agente detect come prima cosa controlla lo stato delle celle (9 in tutto al massimo: 10 contando la poszione stessa in cui si trova l'agente) adiacenti alla sua posizione e invia le informazioni agli altri agenti
                //utilizzando il tc radioChannel
                
                //coordinata x in cui si trova momentaneamente l'agente
                int x=this.getX();

                //coordinata x in cui si trova momentaneamente l'agente
                int y=this.getY();
                
                //emette l'informazione di cella libera relativamente alla cella da lui occupata
                out(tcRadioChannel,new LogicTuple("free_tmp",new Value(x),new Value(y),new Value(""+this.getId())));
                
                
                Struct list=new Struct();
                alice.tuprolog.Var target=new alice.tuprolog.Var("Target");
                
                LogicTuple yetFree;
                Var agName=new Var("AgName");
                
                //scansione dell'indice di colonna (longitudine) del campo minato
                for(int i=x-1;i<=x+1;i++){
                    
                    //scansione dell'indice di riga (latitudine) del campo minato
                    for(int j=y-1;j<=y+1;j++){
                        
                        //variabile utilizzata per essere unificata con il valore rappresentante l'oggetto eventualmente contenuto in una cella
                        Var obj=new Var("Obj");
                        
                        //controllo la cella relativa nel centro di tuple: se la tupla cell(X,Y,Object) non  presente, significa che si tratta di una cella priva di ogni ostacolo
                        LogicTuple res=rdp(tcMineField,new LogicTuple("cell",new Value(i),new Value(j),obj));
                        
                        //se la tupla logica  null allora significa che la cella controllata  free da ostacoli: l'agente inserisce tale  informazione nel centro di tuple radioChannel
                        if(res==null){
                            
                            
                            
                            //prima di inserire la tupla controllo che gi un altro agente non l'abbia dichiarata libera 
                            yetFree=rdp(tcRadioChannel,new LogicTuple("free",new Value(i),new Value(j),agName));
                            
                            //se esiste un risultato non si inserisce la tupla
                            if(yetFree==null)
                                out(tcRadioChannel,new LogicTuple("free_tmp",new Value(i),new Value(j),new Value(""+this.getId())));
                            
                            //lista di un elemento da aggiungere alla lista finora trovata
                            Term[] listToAdd=new Term[1];
                            
                            
                            if( i>=0 && i<xDim && j>=0 && j<yDim ){
                                
                                listToAdd[0]=new LogicTuple("free",new Value(i),new Value(j)).toTerm();
                                //aggiunta della tupla free trovata alla lista che servir per far pianificare all'agente la sua azione
                                //System.out.println(""+new Struct("append",new Struct(listToAdd),list,target));

                                SolveInfo newSol;
                                
                                //if(yetFree==null){
                                     if(this.movDir==1) newSol=engine.solve(new Struct("append",new Struct(listToAdd),list,target));
                                     else newSol=engine.solve(new Struct("append",list,new Struct(listToAdd),target));
                                
                                     list=(Struct)newSol.getTerm("Target");
                                /*}else if(agName.getVarValue("AgName").toString().equals(this.getId().toString())){
                                    System.out.println("TRUEEEEEEEEEEEEEEEEEEEEEEE");
                                    newSol=engine.solve(new Struct("append",new Struct(listToAdd),list,target));
                                     list=(Struct)newSol.getTerm("Target");
                                }*/
                                
                                if(this.matrix[j][i]==-1) this.matrix[j][i]=3;
                            }//if
                            
                        }else{//caso in cui  stata rilevata una mina o un agente
                            
                            //System.out.println(""+obj.getVarValue("Obj"));
                            //se  stata rilevata una mina
                            if( obj.getVarValue("Obj").toString().equals("mine") ){
                                
                               // System.out.println("entrato1");
                                //comunico la cosa agli altri agenti tramite il tc radioChannel: il value none indica che (per quanto riguarda questo agent, nessun agente sminatore  ancora stato
                                //assegnato al disinnesco della mina individuata)
                                out(tcRadioChannel,new LogicTuple("mine_tmp",new Value(i),new Value(j),new Value("none")));
                                
                                //aggiorno l'interfaccia grafica
                                if(this.matrix[j][i]==0) this.matrix[j][i]=5;
                                
                                //lista di un elemento da aggiungere alla lista finora trovata
                                Term[] listToAdd=new Term[1];
                            
                                listToAdd[0]=new LogicTuple("mine",new Value(i),new Value(j)).toTerm();
                                //aggiunta della tupla free trovata alla lista che servir per far pianificare all'agente la sua azione
                                //System.out.println(""+new Struct("append",new Struct(listToAdd),list,target));

                                SolveInfo newSol=engine.solve(new Struct("append",new Struct(listToAdd),list,target));

                                list=(Struct)newSol.getTerm("Target");
                                
                            }else{//controllo se la posizione  occupata da un agente
                                /* 
                                Term found=obj.toTerm();
                                //se il termine trovato unifica con il template di un agente allora lo si aggiunge
                                if(found.unify(new LogicTuple("agent",new Var("X"),new Var("Y"),new Var("Type"),new Var("AgentId")).toTerm())){
                                    
                                    //lista di un elemento da aggiungere alla lista finora trovata
                                    Term[] listToAdd=new Term[1];

                                    listToAdd[0]=found;
                                    //aggiunta della tupla free trovata alla lista che servir per far pianificare all'agente la sua azione
                                    System.out.println(""+new Struct("append",new Struct(listToAdd),list,target));

                                    SolveInfo newSol=engine.solve(new Struct("append",new Struct(listToAdd),list,target));

                                    list=(Struct)newSol.getTerm("Target");
                                    
                                }//if
                                */
                            }//else
                        }//else
                    }//for
                    
                }//for
                
                /*
                //lista di un elemento da aggiungere alla lista finora trovata
                Term[] listToAdd=new Term[1];
                     
                //aggiunta della posizione dell'agente alla lista
                listToAdd[0]=new LogicTuple("agent_me",new Value(this.getX()),new Value(this.getY())).toTerm();
                //aggiunta della tupla free trovata alla lista che servir per far pianificare all'agente la sua azione
                System.out.println(""+new Struct("append",new Struct(listToAdd),list,target));

                SolveInfo newSol=engine.solve(new Struct("append",new Struct(listToAdd),list,target));

                list=(Struct)newSol.getTerm("Target");
                */
                //stampo la lista x debug
               // System.out.println(""+list);
                
                //la lista trovata in precedenza rappresenta lo stato del mondo attorno all'agente detect considerato
                
                //trovo la nuova posizione in cui si muover l'agente
                SolveInfo newSol=engine.solve(new Struct("doAction",list));
                
                //se la sol ha avuto successo (solo per testing)
                if(newSol.isSuccess()){
                    
                    alice.tuprolog.Var varX=new alice.tuprolog.Var("X");
                    alice.tuprolog.Var varY=new alice.tuprolog.Var("Y");
                    alice.tuprolog.Var varType=new alice.tuprolog.Var("Type");
                    
                    //recupero la nuova pos dell'agente e faccio una out corrispondente
                    newSol=engine.solve(new Struct("agentNewPos",varX,varY,varType));
                    
                    //soluzione 
                    
                    Struct solution=(Struct)newSol.getSolution();
                    /*
                    System.out.println(""+solution);
                    System.out.println(""+newSol.getSolution()+";"+varX.getTerm()+","+varY.getTerm());
                    */
                    
                    int oldX=this.getX();
                    int oldY=this.getY();
                    
                    int newX=((Number)solution.getTerm(0)).intValue();
                    int newY=((Number)solution.getTerm(1)).intValue();
                    
                    //cancello il precedente fatto agentNewPos dalla teoria prolog
                    engine.solve("retract(agentNewPos(_,_,_)).");
                    
                    //System.out.println("FACCIO: "+this.getX()+","+this.getY());
                    //aggiorno il centro di tuple con la nuova posizione dell'agente
                    if(rdp(tcRadioChannel,new LogicTuple("agent",new Value(newX),new Value(newY),new Var("Type"),new Var("Name")))==null){
                        
                        
                        //aggiorno la pos corrente dell'agente
                        this.setX(newX);
                        this.setY(newY);
                     
                        
                        
                        engine.solve("update(from("+oldX+","+oldY+"),to("+this.getX()+","+this.getY()+"),"+((Number)solution.getTerm(2)).intValue()+").").getSolution();
                        
                        //resetto la precedente pos dell'agente dall'array per l'aggiornamento grafico
                        this.matrix[oldY][oldX]=3;
                        
                        //nuova pos dell'agente sull'interfaccia grafica
                        this.matrix[this.getY()][this.getX()]=1;
                        
                    }else{//la cella target del movimento  occupata da un altro agente, inverto la direzione e mi muovo in un'altra cella
                        
                            //se non ho raggiunto lo 0 relativamente ai turni di attesa senza movimento aspetto
                            if(waitTurn>0){
                                //System.out.println("Agente detect attende");
                                
                                waitTurn--;
                            }else{
                                engine.solve("changeDirection.");

                                //controllo qual' ora la nuova direzione
                                if(engine.solve("backDirection(down).").isSuccess()){ //direzione di cambio  quella di arretramento sul proprio percorso precedente

                                    //si recupera la cella del movimento nuovo tra quelle visitate
                                    SolveInfo cell=engine.solve("visited("+this.getX()+","+this.getY()+",prev(X1,Y1),Next).");

                                    if(cell.isSuccess()){
                                        
                                        engine.solve("move(from("+this.getX()+","+this.getY()+"),to("+((Number)cell.getTerm("X1")).intValue()+","+((Number)cell.getTerm("Y1")).intValue()+"),1).");
                                        
                                    }else waitTurn--;
                                    

                                }else{//si va in avanti
                                    
                                    //si recupera la cella del movimento nuovo tra quelle visitate
                                    SolveInfo cell=engine.solve("visited("+this.getX()+","+this.getY()+",_,Next).");

                                    if(cell.isSuccess()){
                                        
                                        SolveInfo next=engine.solve(new Struct("element",new Int(1),cell.getTerm("Next"),new Struct("next",new alice.tuprolog.Var("X1"),new alice.tuprolog.Var("Y1"))));
                                        
                                        if(next.isSuccess() ) engine.solve("move(from("+this.getX()+","+this.getY()+"),to("+((Number)next.getTerm("X1")).intValue()+","+((Number)next.getTerm("Y1")).intValue()+"),2).");
                                        else waitTurn--;
                                    }else waitTurn--;
                                    
                                }//else

                                engine.solve("doAction([newPos]).");

                                //recupero la nuova posizione dell'agente
                                SolveInfo newPos=engine.solve("agentNewPos(X,Y,MovType).");

                                //cancello il precedente fatto agentNewPos dalla teoria prolog
                                engine.solve("retract(agentNewPos(_,_,_)).");
                                
                                newX=((Number)newPos.getTerm("X")).intValue();
                                newY=((Number)newPos.getTerm("Y")).intValue();
                                
                                if(rdp(tcRadioChannel,new LogicTuple("agent",new Value(newX),new Value(newY),new Var("Type"),new Var("Name")))==null){
                                    
                                    

                                    //aggiorno la pos corrente dell'agente se nella nuova non  presente a sua volta un agente

                                    this.setX(newX);
                                    this.setY(newY);



                                    engine.solve("update(from("+oldX+","+oldY+"),to("+this.getX()+","+this.getY()+"),"+((Number)solution.getTerm(2)).intValue()+").").getSolution();

                                    //resetto la precedente pos dell'agente dall'array per l'aggiornamento grafico
                                    this.matrix[oldY][oldX]=3;
                                    //nuova pos dell'agente sull'interfaccia grafica
                                    this.matrix[this.getY()][this.getX()]=1;
                                    
                                    waitTurn=5;
                                    
                                }else waitTurn--;
                        }//else
                    }
                }//else System.out.println("fallimento");
                
             
                //System.out.println("FACCIO: "+this.getX()+","+this.getY());
                
                out(tcMineField,new LogicTuple("agent",new Value(this.getX()),new Value(this.getY()),new Value("detect"),new Value(this.getId().toString())));
                
                //l'agente ha finito il proprio turno: lo comunica agli altri agenti: emissione di turn(TurnNum,AgentNum+1) 
                out(tcMineField,new LogicTuple("turn",turnNum.getVarValue("TurnNum"),new Value(agentNo.getVarValue("AgNum").intValue()+1)));
            }//while
            
            
            //fine turno dell'agente e l'agente termina perch  giunta la fine della simulazione 
            out(tcMineField,new LogicTuple("agentEnded",new Value(this.getId().toString())));
            
            
            System.out.println("Agente detect "+this.getId()+" TERMINATO");
            
            
        }catch(Exception e){//nel caso di eccezioni, l'agente sminatore termina il suo compito
            
            System.out.println("DetectAgent "+this.getId()+": terminato a causa di problemi");
            
            e.printStackTrace();
        }
        
    }//body
    
}
