/*
 * UnMineAgent.java
 *
 * Created on 1 dicembre 2004, 19.26
 */

package sid.agent;

import alice.tucson.api.*;
import alice.logictuple.*;

import alice.tuprolog.Prolog;
import alice.tuprolog.Theory;
import alice.tuprolog.Term;
import alice.tuprolog.Number;
import alice.tuprolog.Int;
import alice.tuprolog.Struct;
import alice.tuprolog.SolveInfo;
import java.io.*;

/**
 *Agente sminatore
 * @author  matteo
 */
public class UnMineAgent extends sid.agent.RobotAgent {
    
    
    /**
     *percorso per le teorie per il motore prolog dell'agente
     */
     //private final String thPathCommon=sid.gui.SimulationGUI.userPath+sid.gui.SimulationGUI.fileSeparator+"theories"+sid.gui.SimulationGUI.fileSeparator+"commonTheory.pl";
     
     private final String thPathAgent=sid.gui.SimulationGUI.userPath+sid.gui.SimulationGUI.fileSeparator+"theories"+sid.gui.SimulationGUI.fileSeparator+"agentUnMineTheory.pl";
    
      /**
     *riferimento alla matrice usata per aggiornare l'interfaccia grafica del campo minato
     */
    private int[][] matrix;
    
    
     
    /** 
     *costruttore
     *@param x la coordinata x che relativa alla posizione iniziale dell' Agent
     *@param y la coordinata y relativa alla posizione iniziale dell' Agent
     *@param matrix riferimento alla matrice usata per aggiornare l'interfaccia grafica del campo minato
     *@param name il nome del robot agent, ovvero il suo identificativo
     *
     *
     */
    public UnMineAgent(int x, int y,int[][] matrix, AgentId name) throws Exception {
        
        super(x,y,name);
        
        this.matrix=matrix;
      
        
        //System.out.println(""+unminedMines);
        //System.out.println(""+this.unminedMines);
        
    }//costructor
    
    
    
    /**
     *definizione del body dell'agente sminatore
     *
     */
    protected void body(){
        try{
            
            
            
            
            //creo gli id dei due centri di tuple principali che utilizzer

            //creo il tuplecenter principale che rappresenta il campo minato
            TupleCentreId tcMineField=new TupleCentreId("mineField");


            //creo il tuplecenter principale che rappresenta il campo minato
            TupleCentreId tcRadioChannel=new TupleCentreId("radioChannel");
            
            System.out.println("Agente unMine "+this.getId()+" in esecuzione");
            
            
            //creo un nuovo motore prolog per l'attivita di pianificazione e deliberazione dell'agente
            Prolog engine=new Prolog();
            
            //file con teoria prolog di utilit generale
            //FileInputStream thFileCommon=new FileInputStream(thPathCommon);
            
            //teoria creata a partire dal file precedente
            //Theory commonTheory=new Theory(thFileCommon);
            
            //file con teoria prolog specifica dell'agente
            FileInputStream thFileAgent=new FileInputStream(thPathAgent);
            
            //teoria creata a partire dal file precedente
            Theory agentTheory=new Theory(thFileAgent);
            
            //aggiunta teoria all'engine
            //engine.addTheory(commonTheory);
            engine.addTheory(agentTheory);
             
                   
            
             
             //viene impostata la posizione dell'agente anche nel centro di tuple che rappresenta l'ambiente ovvero il campo minato
             //cn.out(tcMineField,new LogicTuple("cell",new Value(this.getX()),new Value(this.getY()),new Value("agent",new Value("unMine"),new Value(this.getId().toString()))));
             
            
            //appena creato l'agente inserisce la sua posizione nel centro di tuple che rappresenta il canale radio: comunica agli altri la sua posizione
            out(tcRadioChannel,new LogicTuple("agent",new Value(this.getX()),new Value(this.getY()),new Value("unMine"),new Value(this.getId().toString())));
            
            //comunico tramite canale radio che la cella da me occupata  libera
            out(tcRadioChannel,new LogicTuple("free_tmp",new Value(this.getX()),new Value(this.getY()),new Value(""+this.getId())));
            
            //aggiungo la posizione iniziale dell'agente alla teoria e altri fatti utili all'agente per decidere se esite un tragitto per una certa mina
            
            //System.out.println("myPos("+this.getX()+","+this.getY()+"). \n tmpPos("+this.getX()+","+this.getY()+"). \n piano([a,b,c]).");
            
            engine.addTheory(new Theory("myPos("+this.getX()+","+this.getY()+"). \n tmpPos("+this.getX()+","+this.getY()+"). \n piano([]). \n visited("+this.getX()+","+this.getY()+",none,[])."));
            
            
            //l'agente recupera le dimensioni massime del campo minato
            Var m=new Var("MaxDim");
            
            rdp(tcMineField,new LogicTuple("maxDim",new Value("x"),m));
            
            int xDim=m.getVarValue("MaxDim").intValue();

            rdp(tcMineField,new LogicTuple("maxDim",new Value("y"),m));
            
            int yDim=m.getVarValue("MaxDim").intValue();
            
            

            //serve per controllare (ed evitare) movimenti loop dell'agente attorno alle stesse 2 celle
            boolean controlLoop=false;
            
            //turni che l'agente attende senza avere mine vicine da controllare, prima di recuperare anche le mine pi lontane
            int waitTurns=20;
            
            //memorizzano la prima posizione dell'agente
            int xFirstPos=this.getX();
            int yFirstPos=this.getY();
            
            //indica se al turno precedente l'agente ha disinnescato una mine
            boolean mineUnmined=false;
            
            //l'agente cicla fino a che non arriva il termine della simulazione
            while(true){
                
                
                Var agentNo=new Var("AgNum");
                Var turnNum=new Var("TurnNum");
               
                
                in(tcMineField,new LogicTuple("turn",turnNum,agentNo));
                System.out.println("**************************turno numero "+turnNum.getVarValue("TurnNum").intValue()+" di UnMineAgent "+this.getId()+"; agente no: "+agentNo.getVarValue("AgNum"));
                
                //l'agente controlla che non sia arrivata la fine della simulazione
                LogicTuple endSim=rdp(tcMineField,new LogicTuple("endSim"));
                
                //se  arrivata la fine della simulazione allora si esce da ciclo e l'agente termina
                if(endSim!=null) break;
                /*
                //*************************************PER DEBUG CONSULTO LA TEORIA******************************************************************************************
                System.out.println("++++++++++++++++++++++DEBUG "+this.getId()+"++++++++++++++++++++++++++++++++++++");
                SolveInfo debug=engine.solve("myPos(X,Y).");
                System.out.println("mypos: "+debug.getSolution());
                while(engine.hasOpenAlternatives()){
                    
                    System.out.println("mypos: "+engine.solveNext());
                    
                }//while
                debug=engine.solve("tmpPos(X,Y).");
                System.out.println("tmppos: "+debug.getSolution());
                while(engine.hasOpenAlternatives()){
                    
                    System.out.println("tmppos: "+engine.solveNext());
                    
                }//while
                
                debug=engine.solve("initialState(L).");
                if(debug.isSuccess())
                    System.out.println("initialState(L): "+debug.getSolution());
                    while(engine.hasOpenAlternatives()){

                        System.out.println("initialState(L): "+engine.solveNext());

                    }//while
                
                debug=engine.solve("actualDist(D).");
                if(debug.isSuccess())
                    System.out.println("actualDist(D).: "+debug.getSolution());
                    while(engine.hasOpenAlternatives()){

                        System.out.println("actualDist(D).: "+engine.solveNext());

                    }//while
                
                debug=engine.solve("piano(S).");
                System.out.println("piano(S).: "+debug.getSolution());
                while(engine.hasOpenAlternatives()){
                    
                    System.out.println("piano(S).: "+engine.solveNext());
                    
                }//while
                
                debug=engine.solve("visited(X,Y,L,N).");
                System.out.println("visited(X,Y,L,N).: "+debug.getSolution());
                while(engine.hasOpenAlternatives()){
                    
                    System.out.println("visited(X,Y,L,N).: "+engine.solveNext());
                    
                }//while
                System.out.println("+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++");
                */
                //recupero il piano dell'agente dalla teoria prolog
                alice.tuprolog.Var planList= new alice.tuprolog.Var ("L");
                
                SolveInfo plan=engine.solve(new Struct("piano",planList));
                
                //System.out.println(""+plan.getTerm("L"));
                
                //System.out.println(""+(plan.getTerm("L").toString().equals("[]") ? "true" : "false"));
                
                boolean pianifica=plan.getTerm("L").toString().equals("[]");
                
                //se esiste un piano, allora si controlla che la mina selezionata per la cancellazione sia ancora presente (non gi disinnescata da qualche altro agente 
                if(!pianifica){
                    
                 
                    Term actualPlan=plan.getTerm("L");
                    SolveInfo actualMine=engine.solve(new Struct("length",actualPlan,new alice.tuprolog.Var("Num")));
                    actualMine=engine.solve(new Struct("element",actualMine.getTerm("Num"),actualPlan,new Struct("mine", new alice.tuprolog.Var("X"),new alice.tuprolog.Var("Y"),new Struct("none"))));
                    
                    int xMine=((Number)actualMine.getTerm("X")).intValue();
                    int yMine=((Number)actualMine.getTerm("Y")).intValue();
                    
                    //controllo nel radioChannel se la mina  ancora presente
                    LogicTuple mineYetPresent=rdp(tcRadioChannel,new LogicTuple("mine",new Value(xMine),new Value(yMine),new Var("AssignedAgent")));
                    
                    //se presente non faccio nulla, altrimenti cancello il piano attualmente in esecuzione e pianifico per una nuova mina
                    if(mineYetPresent==null){
                        
                        engine.solve("retract(piano(_)).");
                        engine.solve("assert(piano([])).");
                        pianifica=true;
                        System.out.println("MINA ASSEGNATA NON PI PRESENTE: CAMBIO PIANO");
                        
                    }//if
                }//if
                
                //se il piano  vuoto, allora l'agente recupera la pos di tutte le mine segnalate e prova a vedere se esiste un percorso che collega la sua attuale posizione a quella della mina,
                //altrimenti esegue il piano gi in essere
                if( pianifica ){ //pianificazione
                    
                /*
                    //*******************************************************************************************************************************
                    //l'agente recupera dal tc radioChannel le indicazioni di tutte le celle segnalate come free

                    //var list che contiene tutte le tuple free()...
                    Var freeList=new Var("AllFreeCell");

                    inp(tcRadioChannel,new LogicTuple("rd_all",new Value("free",new Var("X"),new Var("Y"),new Var("DetectAgent")),freeList));

                    System.out.println(freeList.toTerm().toString());
                    
                    //*******************************************************************************************************************************
                    
                    //*******************************************************************************************************************************
                    //l'agente recupera dal tc radioChannel le indicazioni di tutte le celle segnalate come MINATE e ancora non assegnate ad alcun agente

                    //var list che contiene tutte le tuple free()...
                    Var mineList=new Var("AllMinedCell");

                   inp(tcRadioChannel,new LogicTuple("rd_all",new Value("mine",new Var("X"),new Var("Y"),new Value("none") ),mineList));

                    System.out.println(mineList.toTerm().toString());
                    //*******************************************************************************************************************************
                    */
                    //conversione delle 2 liste in termini prolog utilizzabili con l'engine tuprolog
                    //Term freeTerm=freeList.toTerm();
                    //Term mineTerm=mineList.toTerm();
                    
                    //alice.tuprolog.Var actualMineX= new alice.tuprolog.Var("MineX");
                    //alice.tuprolog.Var actualMineX= new alice.tuprolog.Var("MineX");
                    
                    Term freeTerm=new Struct();
                    Term mineTerm=new Struct();
                    
                    //si recuperano le caselle libere (da radiochannel) il uno spazio di 49 celle^2 attorno all'agente
                    for(int i=this.getX()-3;i<=this.getX()+3;i++){
                        
                        for(int j=this.getY()-3;j<=this.getY()+3;j++){
                            
                            LogicTuple res=rdp(tcRadioChannel,new LogicTuple("free",new Value(i),new Value(j),new Var("AgName")));
                            
                            //controllo che questa cella dichiarata libera non sia attualmente occupata da nessun altro agente
                            LogicTuple agentInFreeCell=rdp(tcRadioChannel,new LogicTuple("agent",new Value(i),new Value(j),new Var("AgType"),new Var("AgName")));
                            
                            //System.out.println("Agente in freeCell: "+agentInFreeCell+" |");
                            
                            //aggiungo la lista dichiarata libera a quelle fino ad ora recuperate
                            if(res!=null && agentInFreeCell==null){
                                
                                Term[] toAdd=new Term[1];
                                toAdd[0]=res.toTerm();
                                
                                SolveInfo nextList=engine.solve(new Struct("append",freeTerm,new Struct(toAdd),new alice.tuprolog.Var("NextList")));
                                
                                //debug
                                //System.out.println("nuovo append "+nextList);
                                
                                //aggiorno la lista
                                freeTerm=nextList.getTerm("NextList");
                                
                            }//if
                                
                            
                        }//for
                        
                    }//for
                    
                    //si recuperano le caselle minate non ancora assegnate ad un altro agente sminatore (da radiochannel) in uno spazio di 25 celle^2 attorno all'agente
                    
                    //recupero prima le mine adiacenti
                    for(int i=this.getX()-1;i<=this.getX()+1;i++){
                        
                        for(int j=this.getY()-1;j<=this.getY()+1;j++){
                            
                            //System.out.println(""+new LogicTuple("mine",new Value(i),new Value(j),new Value("none")));
                            
                            LogicTuple res=rdp(tcRadioChannel,new LogicTuple("mine",new Value(i),new Value(j),new Value("none")));
                            
                            //aggiungo la lista dichiarata libera a quelle fino ad ora recuperate
                            if(res!=null){
                                
                                Term[] toAdd=new Term[1];
                                toAdd[0]=res.toTerm();
                                
                                SolveInfo next=engine.solve(new Struct("append",mineTerm,new Struct(toAdd),new alice.tuprolog.Var("NextList")));
                                
                                //debug
                                //System.out.println("nuovo append x le mine "+next);
                                
                                //aggiorno la lista
                                mineTerm=next.getTerm("NextList");
                                
                            }//if
                                
                            
                        }//for
                        
                    }//for
                    
                    
                    
                    
                    //colonna che si trova a x+2 rispetto alla pos dell'agente (x)
                    for(int i=this.getY()-1;i<=this.getY()+1;i++){
                        
                        
                            
                            //System.out.println(""+new LogicTuple("mine",new Value(i),new Value(j),new Value("none")));
                            
                            LogicTuple res=rdp(tcRadioChannel,new LogicTuple("mine",new Value(this.getX()+2),new Value(i),new Value("none")));
                            
                            //aggiungo la lista dichiarata libera a quelle fino ad ora recuperate
                            if(res!=null){
                                
                                Term[] toAdd=new Term[1];
                                toAdd[0]=res.toTerm();
                                
                                SolveInfo next=engine.solve(new Struct("append",mineTerm,new Struct(toAdd),new alice.tuprolog.Var("NextList")));
                                
                                //debug
                                //System.out.println("nuovo append x le mine "+next);
                                
                                //aggiorno la lista
                                mineTerm=next.getTerm("NextList");
                                
                            }//if
                                
                            
                        
                        
                    }//for
                    
                    //colonna che si trova a x-2 rispetto alla pos dell'agente (x)
                    for(int i=this.getY()-1;i<=this.getY()+1;i++){
                        
                        
                            
                            //System.out.println(""+new LogicTuple("mine",new Value(i),new Value(j),new Value("none")));
                            
                            LogicTuple res=rdp(tcRadioChannel,new LogicTuple("mine",new Value(this.getX()-2),new Value(i),new Value("none")));
                            
                            //aggiungo la lista dichiarata libera a quelle fino ad ora recuperate
                            if(res!=null){
                                
                                Term[] toAdd=new Term[1];
                                toAdd[0]=res.toTerm();
                                
                                SolveInfo next=engine.solve(new Struct("append",mineTerm,new Struct(toAdd),new alice.tuprolog.Var("NextList")));
                                
                                //debug
                                //System.out.println("nuovo append x le mine "+next);
                                
                                //aggiorno la lista
                                mineTerm=next.getTerm("NextList");
                                
                            }//if
                                
                            
                        
                        
                    }//for
                    
                    
                    
                    
                    //recupero le mine che si trovano nel "perimetro" adiacente a quello realmente adiacente all'agente 
                    
                    //riga che si trova a y-2 rispetto alla pos dell'agente (y)
                    for(int i=this.getX()-2;i<=this.getX()+2;i++){
                        
                        
                            
                            //System.out.println(""+new LogicTuple("mine",new Value(i),new Value(j),new Value("none")));
                            
                            LogicTuple res=rdp(tcRadioChannel,new LogicTuple("mine",new Value(i),new Value(this.getY()-2),new Value("none")));
                            
                            //aggiungo la lista dichiarata libera a quelle fino ad ora recuperate
                            if(res!=null){
                                
                                Term[] toAdd=new Term[1];
                                toAdd[0]=res.toTerm();
                                
                                SolveInfo next=engine.solve(new Struct("append",mineTerm,new Struct(toAdd),new alice.tuprolog.Var("NextList")));
                                
                                //debug
                                //System.out.println("nuovo append x le mine "+next);
                                
                                //aggiorno la lista
                                mineTerm=next.getTerm("NextList");
                                
                            }//if
                                
                            
                        
                        
                    }//for

                    //riga che si trova a y+2 rispetto alla pos dell'agente (y)
                    for(int i=this.getX()-2;i<=this.getX()+2;i++){
                        
                        
                            
                            //System.out.println(""+new LogicTuple("mine",new Value(i),new Value(j),new Value("none")));
                            
                            LogicTuple res=rdp(tcRadioChannel,new LogicTuple("mine",new Value(i),new Value(this.getY()+2),new Value("none")));
                            
                            //aggiungo la lista dichiarata libera a quelle fino ad ora recuperate
                            if(res!=null){
                                
                                Term[] toAdd=new Term[1];
                                toAdd[0]=res.toTerm();
                                
                                SolveInfo next=engine.solve(new Struct("append",mineTerm,new Struct(toAdd),new alice.tuprolog.Var("NextList")));
                                
                                //debug
                                //System.out.println("nuovo append x le mine "+next);
                                
                                //aggiorno la lista
                                mineTerm=next.getTerm("NextList");
                                
                            }//if
                                
                            
                        
                        
                    }//for
                    
                    
                    if(mineTerm.toString().equals("[]") && xFirstPos!=this.getX() && yFirstPos!=this.getY() ){
                        waitTurns--;
                        
                        //se ho aspettato il numero max di turni, allora recupero tutte le mine non assegnate, anche se non adiacenti a me e anche tutte le celle libere
                        if(waitTurns==0){
                            
                            //var list che contiene tutte le tuple free()...
                            Var mineList=new Var("AllMinedCell");

                           inp(tcRadioChannel,new LogicTuple("rd_all",new Value("mine",new Var("X"),new Var("Y"),new Value("none") ),mineList));
                           
                           mineTerm=mineList.toTerm();
                           
                           Var freeList=new Var("AllFreeCell");

                           inp(tcRadioChannel,new LogicTuple("rd_all",new Value("free",new Var("X"),new Var("Y"),new Var("DetectAgent")),freeList));
                           
                           freeTerm=freeList.toTerm();
                           //settaggio di waitTurns al valore iniziale
                           waitTurns=20;
                            
                            
                        }//if
                        
                    }//if
                    
                    //elimino,se presenta, la casella dichiarata libera corrispondente alla posizione attuale dell'agente
                    SolveInfo freeCellNew=engine.solve(new Struct("delete",new Struct("free",new Int(this.getX()),new Int(this.getY()),new alice.tuprolog.Var("_")),freeTerm, new alice.tuprolog.Var("NewFreeList") ));
                    
                    freeTerm=freeCellNew.getTerm("NewFreeList");
                    
                    //ripeto fino a che ci sono mine cio fino a che il goal element.. ha successo
                    boolean isSuccess=true;
                    
                    while(isSuccess){
                        
                        //System.out.println(""+mineTerm);
                        //recupero la prima mina dalla lista delle mine: e cancello il primo elemento
                        SolveInfo mine =engine.solve(new Struct("element",new Int(1),mineTerm,new Struct("mine",new alice.tuprolog.Var("Xm"),new alice.tuprolog.Var("Ym"),new Struct("none") )));
                    
                       // System.out.println("mina sotto esame"+mine);
                        
                        isSuccess=mine.isSuccess();
                        
                        //se non ci sono pi elementi la risoluzione precedente fallisce e quindi si esce dal ciclo
                        if(!isSuccess) break;
                        
                        
                        //resetto alla situazione di base da cui partir l'agente per la sua pianificazione.
                        engine.solve("retract(tmpPos(_,_)).");
                        engine.solve("retract(myPos(_,_)).");
                        engine.solve("assert(myPos("+this.getX()+","+this.getY()+")).");
                        engine.solve("assert(tmpPos("+this.getX()+","+this.getY()+")).");
                        engine.solve("retract(initialState(_)).");
                        engine.solve("retract(actualDist(_)).");
                        engine.solve("retract(piano(_)).");
                        engine.solve("assert(piano([])).");
                        
                        Number xM=((Number)mine.getTerm("Xm"));
                        Number yM=((Number)mine.getTerm("Ym"));
                        
                         //elimino dalla lista la mine appena ottenuta come riferimento per una possibile pianificazione
                         SolveInfo deleteMine =engine.solve(new Struct("delete",new Struct("mine",xM,yM,new alice.tuprolog.Var("_") ),mineTerm, new alice.tuprolog.Var("NewMineList") ));
                    
                         //System.out.println("mina sotto esame cancellata"+deleteMine);
                        
                         //riassegnazione della nuova lista di mine
                         mineTerm=deleteMine.getTerm("NewMineList");
                         
                        //********************************ricavo eventuale piano per la mina considerata********************************************************************************
                        
                        
                        //inserisco nella teoria il nuovo fatto initialState, usato nel caso il pianificatore prenda un vicolo cieco: con initialState([allFreeCell]) il pianificatore 
                        //pu tornare indietro e vedere se esiste una strada alternativa a quella "senza uscita" (escudendo chiaramente dallo stato iniziale l'ultima cella libera che ha
                        //decretato il vicolo cieco
                        engine.solve(new Struct("assert",new Struct("initialState",freeTerm)));

                        //asserisco il fatto actualDist(D) che rappresenta la distanza attuale dalla mina rispetto alla posizione attuale dell'agente (distanza "in linea d'aria")
                        engine.solve("assert(actualDist("+sid.util.UnMineUtility.getDist(this.getX(),this.getY(),xM.intValue(),yM.intValue())+")).");
                        
                        //System.out.println("distanza attuale:"+engine.solve("actualDist(D)."));
                        
                        
                        //l'agente controlla se per la mina data e le celle libere esistenti esiste un piano
                        SolveInfo piano=engine.solve(new Struct("plan",freeTerm,new Struct("mine",xM,yM,new Struct("none"))));
                        
                        //System.out.println("piano trovato: "+engine.solve("piano(_).").getSolution());
                        //System.out.println("distanza post piano"+engine.solve("actualDist(D)."));
                        
                        if(!engine.solve("piano(L).").getTerm("L").toString().equals("[]")) {
                            
                            //System.out.println("piano trovato");
                            engine.solve("retract(initialState(_)).");
                            engine.solve("retract(actualDist(D)).");
                            engine.solve("assert(tmpPos("+this.getX()+","+this.getY()+")).");
                            
                            //la mina nel radiochannel risulter ora assegnata a questo agente
                            inp(tcRadioChannel,new LogicTuple("mine",new Value(xM.intValue()),new Value(yM.intValue()),new Value("none") ));
                            
                            //aggiorno l'int grafica
                            this.matrix[yM.intValue()][xM.intValue()]=4;
                            
                            //mina assegnata a questo agente
                            out(tcRadioChannel,new LogicTuple("mine",new Value(xM.intValue()),new Value(yM.intValue()),new Value(""+this.getId()) ));
                            
                            //System.out.println(""+engine.getTheory());
                            break; //trovato un percorso verso la mina
                        }
                        
                        engine.solve("assert(tmpPos("+this.getX()+","+this.getY()+")).");
                        engine.solve("retract(initialState(_)).");
                        engine.solve("retract(actualDist(D)).");
                        
                       // System.out.println(""+engine.getTheory());
                        
                    }//while
                    
                    //se si  usciti dal ciclo per mancanza di ulteriori mine da controllare, allora l'agente si muove in una posizione adiacente, dichiarata libera
                    if(!isSuccess){
                        
                        
                        if(!engine.solve("tmpPos(X,Y).").isSuccess()){
                            
                           // System.out.println("non c' pi tmpPos");
                            engine.solve("assert(tmpPos("+this.getX()+","+this.getY()+")).");
                        }//if
                        
                        //recupero la lista delle celle libere adiacenti alla posizione attuale dell'agente
                        SolveInfo freeAdiacent=engine.solve(new Struct("adiacent",freeTerm,new alice.tuprolog.Var("Adiacent")));
                        
                        Term freeTermAdiacent=freeAdiacent.getTerm("Adiacent");
                        
                        Term initialFreeTermAdiacent=freeTermAdiacent;
                        
                        
                        //recupero la prima cella libera dalla lista 
                        SolveInfo move =engine.solve(new Struct("element",new Int(1),freeTermAdiacent,new Struct("free",new alice.tuprolog.Var("X"),new alice.tuprolog.Var("Y"),new alice.tuprolog.Var("AgName") )));
                    
                        
                        Number xNext=new Int(-1);
                        Number yNext=new Int(-1);
                            
                        boolean freeCellFound=false;
                        boolean doMove=true;
                        
                        while(move.isSuccess()){
                            
                            
                            xNext=((Number)move.getTerm("X"));
                            yNext=((Number)move.getTerm("Y"));
                            
                            //controllo se si tratta di una cella gi visitata
                            SolveInfo yetVisited=engine.solve("visited("+xNext+","+yNext+",_,_).");
                            
                            //se la soluzione  false allora si tratta di uno stato non ancora visitato: diventer la nuova cella in cui si muover l'agente
                            if(!yetVisited.isSuccess()){
                                
                               // System.out.println("trovata cella non ancora visitata");
                                freeCellFound=true;
                                controlLoop=false;
                                break;
                            }//if
                            
                            move=engine.solve(new Struct("delete",new Struct("free",new Int(xNext.intValue()),new Int(yNext.intValue()),new alice.tuprolog.Var("AgName") ),freeTermAdiacent,new alice.tuprolog.Var("NewFreeList")));
                            
                           // System.out.println("move delete: "+move);
                            
                            freeTermAdiacent=move.getTerm("NewFreeList");
                            
                            move =engine.solve(new Struct("element",new Int(1),freeTermAdiacent,new Struct("free",new alice.tuprolog.Var("X"),new alice.tuprolog.Var("Y"),new alice.tuprolog.Var("AgName") )));
                           // System.out.println("move element: "+move);
                            
                        }//while
                       
                        //se tra le celle libere non c' ne  nessuna nn ancora visitata, allora l'agente si muove alla cella precedente rispetto a quella in cui si trova ora
                        if(!freeCellFound){
                            
                            //scelta della direzione nella quale muoversi (tra le celle gi visitate)
                            SolveInfo backDir=engine.solve("backDirection(down).");
                            
                            //se true mi muovo a ritroso
                            if(backDir.isSuccess()){
                                
                                //seleziono la cella precedente rispetto a quella in cui mi trovo
                                SolveInfo prevCell=engine.solve("visited("+this.getX()+","+this.getY()+",precedent(Xp,Yp),L).");
                                
                                SolveInfo next=engine.solve("visited("+this.getX()+","+this.getY()+",none,L).");
                                
                                //se true allora esiste, altrimenti cambio direzione e mi muovo in avanti
                                if(prevCell.isSuccess()){
                                    
                                    xNext=((Number)prevCell.getTerm("Xp"));
                                    yNext=((Number)prevCell.getTerm("Yp"));
                                   
                                    
                                }else if( ! next.getTerm("L").toString().equals("[]") ){
                                    
                                    //cambio direzione
                                    engine.solve("changeDirection.");
                                    
                                    //System.out.println("direzione cambiata? "+engine.solve("backDirection(up)."));
                                    //recupero la cella target del prossimo movimento
                                    SolveInfo nextCell=engine.solve("visited("+this.getX()+","+this.getY()+",Prev,L).");
                                    
                                    Term nextList=nextCell.getTerm("L");
                                    Term prev=nextCell.getTerm("Prev");
                                    
                                    engine.solve("retract(visited("+this.getX()+","+this.getY()+",_,_)).");
                                    
                                    //ottengo il primo elemento della lista
                                    nextCell =engine.solve(new Struct("element",new Int(1),nextList,new Struct("next",new alice.tuprolog.Var("X"),new alice.tuprolog.Var("Y") )));
                                    
                                    
                                    xNext=((Number)nextCell.getTerm("X"));
                                    yNext=((Number)nextCell.getTerm("Y"));
                                                                     
                                   //cambio l'ordine con cui  memorizzata la lista dei next
                                   Term newNextList=engine.solve(new Struct("delete",new Struct("next",new Int(xNext.intValue()),new Int(yNext.intValue())),new alice.tuprolog.Var("NewList"))).getTerm("NewList");
                                   
                                   //aggiungo l'elemento eliminato in fondo
                                   
                                   Term[] toAdd=new Term[1];
                                   
                                   toAdd[0]=new Struct("next",new Int(xNext.intValue()),new Int(yNext.intValue()));
                                   
                                   Term defNextList=engine.solve(new Struct("append",newNextList,new Struct(toAdd),new alice.tuprolog.Var("NewDefList"))).getTerm("NewDefList");
                                   
                                   engine.solve(new Struct("assert",new Struct("visited",new Int(this.getX()),new Int(this.getY()),prev,defNextList)));
                                   
                                    
                                }else doMove=false;
                                
                            }else{
                                
                                    SolveInfo nextCell=engine.solve("visited("+this.getX()+","+this.getY()+",Prev,L).");
                                    
                                    if(nextCell.isSuccess()){
                                        Term nextList=nextCell.getTerm("L");
                                        Term prev=nextCell.getTerm("Prev");

                                        engine.solve("retract(visited("+this.getX()+","+this.getY()+",_,_)).");

                                        //ottengo il primo elemento della lista
                                        nextCell =engine.solve(new Struct("element",new Int(1),nextList,new Struct("next",new alice.tuprolog.Var("X"),new alice.tuprolog.Var("Y") )));


                                        xNext=((Number)nextCell.getTerm("X"));
                                        yNext=((Number)nextCell.getTerm("Y"));

                                       //cambio l'ordine con cui  memorizzata la lista dei next
                                       Term newNextList=engine.solve(new Struct("delete",new Struct("next",new Int(xNext.intValue()),new Int(yNext.intValue())),new alice.tuprolog.Var("NewList"))).getTerm("NewList");

                                       //aggiungo l'elemento eliminato in fondo

                                       Term[] toAdd=new Term[1];

                                       toAdd[0]=new Struct("next",new Int(xNext.intValue()),new Int(yNext.intValue()));

                                       Term defNextList=engine.solve(new Struct("append",newNextList,new Struct(toAdd),new alice.tuprolog.Var("NewDefList"))).getTerm("NewDefList");

                                       engine.solve(new Struct("assert",new Struct("visited",new Int(this.getX()),new Int(this.getY()),prev,defNextList)));
                                    }else{
                                        
                                        //cambio direzione
                                        engine.solve("changeDirection.");
                                        
                                        //seleziono la cella precedente rispetto a quella in cui mi trovo
                                        SolveInfo prevCell=engine.solve("visited("+this.getX()+","+this.getY()+",precedent(Xp,Yp),_).");

                                        
                                         xNext=((Number)prevCell.getTerm("Xp"));
                                         yNext=((Number)prevCell.getTerm("Yp"));

                                      }//else
                            }//else
                            
                            /*
                            if(controlLoop){
                                //vado in una cella "successiva" rispetto a questa
                                
                                SolveInfo nextMove=engine.solve(new Struct("element",new Int(1),initialFreeTermAdiacent,new Struct("free",new alice.tuprolog.Var("X"),new alice.tuprolog.Var("Y"),new alice.tuprolog.Var("AgName") )));
                                xNext=((Number)nextMove.getTerm("X"));
                                yNext=((Number)nextMove.getTerm("Y"));
                                
                                
                                
                            }else{
                                //recupero la cella precedente rispetto a quella in cui mi trovo
                                SolveInfo prevCell=engine.solve("visited("+this.getX()+","+this.getY()+",precedent(Xp,Yp),_).");

                                if(prevCell.isSuccess()){
                                    xNext=((Number)prevCell.getTerm("Xp"));
                                    yNext=((Number)prevCell.getTerm("Yp"));
                                    controlLoop=true;
                                    
                                }else doMove=false;
                            
                            }
                            */
                        }//if
                        /*
                        SolveInfo move=engine.solve(new Struct("moveToFreeCell", freeTerm, new Struct("actual",new Int(this.getX()),new Int(this.getY())),new Struct("res",new alice.tuprolog.Var("X"),new alice.tuprolog.Var("Y"))));
                    
                         Number xNext=((Number)move.getTerm("X"));
                         Number yNext=((Number)move.getTerm("Y"));
                         */
                         //l'agente si muove verso la nuova posizione
                        
                        //prima si controlla che la nuova posizione non sia occupata da un agente
                        //controllo se nel campo minato, nella cella relativa al mio movimento, vi sia un'altro agente
                         LogicTuple fullCell=rdp(tcMineField,new LogicTuple("cell",new Value(xNext.intValue()),new Value(yNext.intValue()),new Value("agent",new Var("Type"),new Var("Name"))));

                         if(fullCell!=null){
                             //l'agente non si muove per questo turno
                             doMove=false;
                            // System.out.println("L'agente aspetta per questo turno");
                         }//if
                             
                        if(doMove){
                            
                            
                            
                             //aggiornamento int. grafica: vecchia cella free (bandiera verde o croce rossa se l'agente  su una pos in precedenza occupata da una mina
                             if(mineUnmined){
                                 
                                 this.matrix[this.getY()][this.getX()]=6;
                                 mineUnmined=false;
                                 
                             }else this.matrix[this.getY()][this.getX()]=3;

                             int oldX=this.getX();
                             int oldY=this.getY();

                             this.setX(xNext.intValue());
                             this.setY(yNext.intValue());
                             
                             
                             

                             if(freeCellFound) {
                                 SolveInfo updatePrevVisited=engine.solve("visited("+oldX+","+oldY+",S,L).");

                                 Term[] toAdd=new Term[1];

                                 Term prev=updatePrevVisited.getTerm("S");

                                 toAdd[0]=new Struct("next",new Int(this.getX()),new Int(this.getY()));

                                 SolveInfo newNext=engine.solve(new Struct("append",updatePrevVisited.getTerm("L"),new Struct(toAdd),new alice.tuprolog.Var("NewList")));

                                 //System.out.println("new Next: "+newNext.getSolution());
                                 engine.solve("retract(visited("+oldX+","+oldY+",precedent(X,Y),_)).");

                                 engine.solve(new Struct("assert",new Struct("visited",new Int(oldX),new Int(oldY),prev,newNext.getTerm("NewList"))));

                                 //engine.solve("retract(visited("+this.getX()+","+this.getY()+",precedent("+oldX+","+oldY+"),_)).");

                                 //l'agente, nella sua teoria prolog tiene traccia dell'intero percorso che ha seguito

                                 engine.solve("assert(visited("+this.getX()+","+this.getY()+",precedent("+oldX+","+oldY+"),[])).");
                             }
                             
                             //aggiornamento int. grafica: nuova cella occupata dall'agente
                             this.matrix[this.getY()][this.getX()]=2;

                             out(tcMineField,new LogicTuple("agent",new Value(this.getX()),new Value(this.getY()),new Value("unMine"),new Value(this.getId().toString())));

                             //aggiorno la teoria prolog
                             engine.solve("retract(myPos(X,Y)).");
                             engine.solve("retract(tmpPos(X,Y)).");
                             engine.solve("assert(myPos("+this.getX()+","+this.getY()+")).");
                             engine.solve("assert(tmpPos("+this.getX()+","+this.getY()+")).");
                        }
                        
                    }//if
                    
                }else{//eseguo il piano gi in essere
                    
                    Term actualPlan=plan.getTerm("L");
                    
                    //System.out.println("piano attualmente in esecuzione "+actualPlan);
                    
                     //recupero la prossima cella nella quale l'agente si deve muovere
                     SolveInfo nextCell =engine.solve(new Struct("element",new Int(1),actualPlan,new Struct("free",new alice.tuprolog.Var("Xm"),new alice.tuprolog.Var("Ym"),new alice.tuprolog.Var("AgName") )));
                    
                     
                     
                     
                     
                     if(nextCell.isSuccess()){
                         
                         Number xNext=((Number)nextCell.getTerm("Xm"));
                         Number yNext=((Number)nextCell.getTerm("Ym"));
                         
                         //controllo se nel campo minato, nella cella relativa al mio movimento, vi sia un'altro agente
                         LogicTuple fullCell=rdp(tcMineField,new LogicTuple("cell",new Value(xNext.intValue()),new Value(yNext.intValue()),new Value("agent",new Var("Type"),new Var("Name"))));

                         if(fullCell!=null){
                             //l'agente non si muove per questo turno
                             //System.out.println("agente UNMINE aspetta");
                         }else{
                             //l'agente si muove verso la nuova posizione

                            //aggiornamento int. grafica: vecchia cella free (bandiera verde o croce rossa se l'agente  su una pos in precedenza occupata da una mina
                            if(mineUnmined){
                                 
                                 this.matrix[this.getY()][this.getX()]=6;
                                 mineUnmined=false;
                                 
                             }else this.matrix[this.getY()][this.getX()]=3;

                             int oldX=this.getX();
                             int oldY=this.getY();
                             
                             this.setX(xNext.intValue());
                             this.setY(yNext.intValue());
                             
                             
                              SolveInfo updatePrevVisited=engine.solve("visited("+oldX+","+oldY+",S,L).");
                             
                             Term[] toAdd=new Term[1];
                             
                             Term prev=updatePrevVisited.getTerm("S");
                             
                             toAdd[0]=new Struct("next",new Int(this.getX()),new Int(this.getY()));
                             
                             SolveInfo newNext=engine.solve(new Struct("append",updatePrevVisited.getTerm("L"),new Struct(toAdd),new alice.tuprolog.Var("NewList")));
                             
                             engine.solve("retract(visited("+oldX+","+oldY+",precedent(X,Y),_)).");
                             
                             engine.solve(new Struct("assert",new Struct("visited",new Int(oldX),new Int(oldY),prev,newNext.getTerm("NewList"))));
                             
                             engine.solve("retract(visited("+this.getX()+","+this.getY()+",precedent("+oldX+","+oldY+"),_)).");

                             //l'agente, nella sua teoria prolog tiene traccia dell'intero percorso che ha seguito
                             engine.solve("assert(visited("+this.getX()+","+this.getY()+",precedent("+oldX+","+oldY+"),[])).");
                             
                             //aggiornamento int. grafica: nuova cella occupata dall'agente
                             this.matrix[this.getY()][this.getX()]=2;

                             out(tcMineField,new LogicTuple("agent",new Value(this.getX()),new Value(this.getY()),new Value("unMine"),new Value(this.getId().toString())));

                              //aggiorno la teoria prolog
                              engine.solve("retract(myPos(X,Y)).");
                              engine.solve("retract(tmpPos(X,Y)).");
                              engine.solve("assert(myPos("+this.getX()+","+this.getY()+")).");
                              engine.solve("assert(tmpPos("+this.getX()+","+this.getY()+")).");
                             
                             //aggiorno il piano anche nella teoria prolog
                             SolveInfo nuovoPiano=engine.solve(new Struct("delete",new Struct("free",new Int(xNext.intValue()),new Int(yNext.intValue()),new alice.tuprolog.Var("AgName") ),actualPlan, new alice.tuprolog.Var("NewPlan") ));
                             //System.out.println(""+engine.solve("retract(piano(L)).")+""+new Struct("updatePlan",nuovoPiano.getTerm("NewPlan")));
                             engine.solve(new Struct("updatePlan",nuovoPiano.getTerm("NewPlan")));
                             //System.out.println("nuovo piano asserito "+nuovoPiano.getTerm("NewPlan"));
                             //System.out.println("piano nuovo "+engine.solve("piano(L)."));
                             
                         }//else
                    }else{//sono arrivato al punto in cui devo disinnescare la mina
                    
                    //recupero la mina
                    SolveInfo finalPos =engine.solve(new Struct("element",new Int(1),actualPlan,new Struct("mine",new alice.tuprolog.Var("Xm"),new alice.tuprolog.Var("Ym"),new Struct("none") )));
                    
                    Number xFin=((Number)finalPos.getTerm("Xm"));
                    Number yFin=((Number)finalPos.getTerm("Ym"));
                    
                    if(mineUnmined){
                        this.matrix[this.getY()][this.getX()]=6;
                        mineUnmined=false;
                    }else this.matrix[this.getY()][this.getX()]=3;
                    
                    //settaggio int  grafica
                    mineUnmined=true;
                    
                    int oldX=this.getX();
                    int oldY=this.getY();
                    
                    
                    
                    this.setX(xFin.intValue());
                    this.setY(yFin.intValue());
                    
                     SolveInfo updatePrevVisited=engine.solve("visited("+oldX+","+oldY+",S,L).");
                             
                     Term[] toAdd=new Term[1];
                             
                     Term prev=updatePrevVisited.getTerm("S");
                             
                     toAdd[0]=new Struct("next",new Int(this.getX()),new Int(this.getY()));
                             
                      SolveInfo newNext=engine.solve(new Struct("append",updatePrevVisited.getTerm("L"),new Struct(toAdd),new alice.tuprolog.Var("NewList")));
                             
                      engine.solve("retract(visited("+oldX+","+oldY+",precedent(X,Y),_)).");
                             
                      engine.solve(new Struct("assert",new Struct("visited",new Int(oldX),new Int(oldY),prev,newNext.getTerm("NewList"))));
                             
                      engine.solve("retract(visited("+this.getX()+","+this.getY()+",precedent("+oldX+","+oldY+"),_)).");
                    
                    
                    //l'agente, nella sua teoria prolog tiene traccia dell'intero percorso che ha seguito
                    engine.solve("assert(visited("+this.getX()+","+this.getY()+",precedent("+oldX+","+oldY+"),[])).");
                             
                    
                    

                    this.matrix[this.getY()][this.getX()]=2;
                    
                    
                    inp(tcMineField,new LogicTuple("mine",new Value(this.getX()),new Value(this.getY()),new Value(""+this.getId())));
                    
                    //aggiorno il numero totale di mine disinnescate
                    out(tcMineField,new LogicTuple("incUnminedMines"));
                    
                    out(tcMineField,new LogicTuple("agent",new Value(this.getX()),new Value(this.getY()),new Value("unMine"),new Value(this.getId().toString())));
                    
                    //aggiorno la teoria prolog
                    engine.solve("retract(myPos(X,Y)).");
                    engine.solve("retract(tmpPos(X,Y)).");
                    engine.solve("assert(myPos("+this.getX()+","+this.getY()+")).");
                    engine.solve("assert(tmpPos("+this.getX()+","+this.getY()+")).");
                    
                    
                    SolveInfo finalPlan=engine.solve(new Struct("delete",new Struct("mine",new Int(xFin.intValue()),new Int(yFin.intValue()),new Struct("none") ),actualPlan, new alice.tuprolog.Var("FinalPlan") ));
                    engine.solve("retract(piano(L)).");
                    engine.solve("assert(piano([])).");
                    
                    System.out.println("piano concluso "+engine.solve("piano(L)."));
                    
                }//else
                
                }//else
                //fine turno dell'agente
                out(tcMineField,new LogicTuple("turn",turnNum.getVarValue("TurnNum"),new Value(agentNo.getVarValue("AgNum").intValue()+1)));
            }//while
            
            
            //fine turno dell'agente e l'agente termina perch  giunta la fine della simulazione 
            out(tcMineField,new LogicTuple("agentEnded",new Value(this.getId().toString())));
            
            System.out.println("Agente unMine "+this.getId()+" TERMINATO");
            
            
        }catch(Exception e){//nel caso di eccezioni, l'agente sminatore termina il suo compito
            
            System.out.println("UnMineAgent "+this.getId()+": terminato a causa di problemi");
            
            e.printStackTrace();
        }
    }//body
    
}
